﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace PowerCodeCalculator.Service
{
    using PowerCodeCalculator.Model;
    using PowerCodeCalculator.Util;

    /// <summary>
    /// 代码函数统计服务
    /// </summary>
    public class CodeCounterService
    {
        #region 属性

        #endregion

        #region 方法

public static CodeStatisticInfo Calc(CodeStatisticCondition condition)
{
    //1、验证参数
    var statResult = CheckCondition(condition);
    if (statResult.IsOK == false)
    {
        return statResult;
    }

    //2、递归获取所有代码目录
    var dirList = GetFileDir(condition.CodeDirectory);

    //3、以目录为单位进行遍历统计
    try
    {
        if (condition.IsEnableMultiThread == false) //单线程统计
        {
            foreach (var dir in dirList)
            {
                var innerCodeInfoList = GetCodeInfoList(dir, condition);
                statResult.CodeInfoList.AddRange(innerCodeInfoList);
            }
        }
        else //多线程统计 待统计代码目录较多，文件较大或者个人PC配置渣渣的情况下请慎用
        {
            var taskCnt = 8; //并发任务数
            var dictDir = GetGroupDir(taskCnt, dirList);
            var taskFactory = new TaskFactory();
            foreach (var kv in dictDir)
            {
                PowerLogger.AppendLog("任务【{0}】,并发执行{1}个目录的代码统计", kv.Key, kv.Value.Count);
                var listTask = new List<Task<IList<CodeInfo>>>();
                foreach (var item in kv.Value)
                {
                    var notifyTask = taskFactory.StartNew<IList<CodeInfo>>
                    (
                        () => { return GetCodeInfoList(item, condition); }
                    );

                    listTask.Add(notifyTask);
                }

                Task.WaitAll(listTask.ToArray()); //等待一批任务并行执行完成

                foreach (var tmpTask in listTask)
                {
                    statResult.CodeInfoList.AddRange(tmpTask.Result);
                }
            }
        }
    }
    catch (Exception ex)
    {
        statResult.IsOK = false;
        statResult.OutPutInfo = string.Format("统计代码发生异常：{0}", ex);
    }
    if (statResult.IsOK == false)
    {
        return statResult;
    }

    //4、统计代码总行数及输出
    statResult.TotalCount = statResult.CodeInfoList.Sum(x => x.LineCount);
    statResult.TotalCharCount = statResult.CodeInfoList.Sum(x => x.CharCount);

    var sb = new StringBuilder(1024);
    sb.AppendFormat("代码总行数:{0}", statResult.TotalCount);
    sb.AppendLine();
    sb.AppendFormat("代码总字符数:{0}", statResult.TotalCharCount);
    sb.AppendLine();

    if (statResult.CodeInfoList.Any())
    {
        sb.Append("代码行数统计明细：");
        sb.AppendLine();
    }
    else
    {
        sb.Append("没有匹配的代码文件");
    }
    foreach (var codeInfo in statResult.CodeInfoList)
    {
        sb.Append(codeInfo.FilePath);
        sb.AppendLine();
        sb.AppendFormat("【代码行数】:{0}", codeInfo.LineCount);
        sb.AppendLine();
        sb.AppendFormat("【代码字符数】:{0}", codeInfo.CharCount);

        sb.AppendLine();
        sb.AppendLine();
    }

    statResult.OutPutInfo = sb.ToString();

    return statResult;
}

        #endregion

        #region 辅助方法

        /// <summary>
        /// 验证条件参数
        /// </summary>
        /// <param name="condition"></param>
        private static CodeStatisticInfo CheckCondition(CodeStatisticCondition condition)
        {
            var statResult = new CodeStatisticInfo
            {
                IsOK = true,
                TotalCount = 0,
                CodeInfoList = new List<CodeInfo>(),
                TotalCharCount = 0,
                OutPutInfo = string.Empty,
            };

            if (condition == null)
            {
                statResult.IsOK = false;
                statResult.OutPutInfo = "统计条件为空";
            }
            else if (string.IsNullOrWhiteSpace(condition.CodeDirectory) == true)
            {
                statResult.IsOK = false;
                statResult.OutPutInfo = "代码目录为空";
            }
            else if (System.IO.Directory.Exists(condition.CodeDirectory) == false)
            {
                statResult.IsOK = false;
                statResult.OutPutInfo = string.Format("{0}目录不存在", condition.CodeDirectory);
            }
            else if (string.IsNullOrWhiteSpace(condition.FileTypeFilter) == true)
            {
                statResult.IsOK = false;
                statResult.OutPutInfo = "统计的代码文件类型为空";
            }

            return statResult;
        }

        /// <summary>
        /// 递归获取文件目录
        /// </summary>
        /// <param name="dirPath"></param>
        /// <returns></returns>
        private static IList<string> GetFileDir(string dirPath)
        {
            var listDir = new List<string> { dirPath };

            var dirs = Directory.GetDirectories(dirPath);
            listDir.AddRange(dirs);

            foreach (var item in dirs)
            {
                var innerDir = GetFileDir(item);
                listDir.AddRange(innerDir);
            }

            listDir = listDir.Distinct().ToList();

            ////忽略统计的文件夹
            //var ignorePathList = new List<string> { "bin", "obj", "Properties" };

            return listDir;
        }

        /// <summary>
        /// 按目录统计代码信息
        /// </summary>
        /// <param name="dirPath"></param>
        /// <param name="condition"></param>
        /// <returns></returns>
        private static IList<CodeInfo> GetCodeInfoList(string dirPath, CodeStatisticCondition condition)
        {
            var codeList = new List<CodeInfo>();
            var filePaths = Directory.GetFiles(dirPath);

            filePaths = filePaths.Where(x => Path.GetExtension(x).ToLower() == condition.FileTypeFilter).ToArray();
            if (string.IsNullOrWhiteSpace(condition.IgnoreFileType) == false)
            {
                var arrIgnoreType = condition.IgnoreFileType.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
                foreach (var ignoreType in arrIgnoreType)
                {
                    filePaths = filePaths.Where(x => x.ToLower().Contains(ignoreType) == false).ToArray();
                }
            }

            if (filePaths.Any() == false)
            {
                return codeList;
            }

            foreach (var tmpPath in filePaths)
            {
                var codeInfo = new CodeInfo
                {
                    FilePath = tmpPath,
                };

                var lines = File.ReadLines(tmpPath, Encoding.UTF8);
                foreach (var tmpLine in lines)
                {
                    var strCode = tmpLine.Trim();
                    if (condition.IsCalcBrace == false) //不统计大括号
                    {
                        if (strCode == "{" || strCode == "}")
                        {
                            continue;
                        }
                    }
                    if (condition.IsCalcEmptyLine == false) //不统计空行
                    {
                        if (strCode.Length == 0)
                        {
                            continue;
                        }
                    }
                    if (condition.IsCalcComment == false) //不统计注释 /**/这种大片段类型的注释暂不考虑
                    {
                        if (strCode.StartsWith(@"//"))
                        {
                            continue;
                        }
                        //sql注释特殊处理
                        if (condition.FileTypeFilter == ".sql" && strCode.StartsWith("--"))
                        {
                            continue;
                        }
                    }

                    //统计代码行数
                    codeInfo.LineCount++;

                    //统计字符数
                    codeInfo.CharCount += strCode.Length;
                }


                codeList.Add(codeInfo);
            }

            return codeList;
        }

        #endregion

        #region 多线程统计帮助方法

        /// <summary>
        /// 对目录分组
        /// </summary>
        /// <param name="taskCnt">每组任务个数</param>
        /// <param name="listAllDir">待处理目录列表</param>
        /// <returns></returns>
        private static Dictionary<int, List<string>> GetGroupDir(int taskCnt, IList<string> listAllDir)
        {
            //将任务分组 n个一组
            var dict = new Dictionary<int, List<string>>();

            var listChildToDoInfo = new List<string>();
            var groupID = 0;//分组编号
            for (int i = 0; i < listAllDir.Count; i++)
            {
                var todoInfo = listAllDir[i];
                listChildToDoInfo.Add(todoInfo);
                if (listChildToDoInfo.Count == taskCnt)
                {
                    dict.Add(groupID, listChildToDoInfo);
                    groupID++;
                    listChildToDoInfo = new List<string>();
                }
                else if (i == listAllDir.Count - 1)
                {
                    dict.Add(groupID, listChildToDoInfo);
                    groupID++;
                }
            }

            return dict;
        }

        #endregion
    }
}
